#####################################################################
##### Multivariate tail coefficients: properties and estimation #####
#####            I. Gijbels, V. Kika, M. Omelka                 #####
#####               Real data application                       #####
#####################################################################


##### set working directory to the source file location

# in RStudio
# setwd(dirname(rstudioapi::getActiveDocumentContext()$path))



# install.packages("dplyr")
# install.packages("rugarch")
# install.packages("qrmtools")
# install.packages("copula")
# install.packages("zoo")
# install.packages("readr")

library(dplyr)
library(rugarch)
library(qrmtools)
library(copula)
library(zoo)
library(readr)


##### List of functions

### Lower coefficient Li
lowerli<-function(data,J=1){ # J - the set of indices to be conditioned on
  n <- nrow(data)
  d<- ncol(data)
  pseudodata<-pobs(data)
  res<-as.data.frame(matrix(NA,ncol=2, nrow=n-1))  # k=1, ..., n-1
  names(res)<-c("k", "lambda")
  res[,1]<-1:(n-1)
  
  Cn <- C.n(matrix((1:(n-1))/(n+1), nrow=n-1, ncol=d), pseudodata)
  Cndenom <- C.n(matrix((1:(n-1))/(n+1), nrow=n-1, ncol=length(J)), as.data.frame(pseudodata[,J]))
  
  res[,2] <- Cn/Cndenom;
  return(res)
  }



### Upper coefficient Li
upperli<-function(data,J=1){ # J - the set of indices to be conditioned on
  n <- nrow(data)
  d<- ncol(data)
  pseudodata<-pobs(data)
  res<-as.data.frame(matrix(NA,ncol=2, nrow=n-1))  # k=1, ..., n-1
  names(res)<-c("k", "lambda")
  res[,1]<-1:(n-1)
  
  pseudodata2 <- 1 - pseudodata + 1/n^2; # the last summand to obtain strict inequality in empirical survival function
  Cn.surv <- C.n(matrix(1-((n-1):1)/(n+1), nrow=n-1, ncol=d), pseudodata2)
  Cn.survdenom <- C.n(matrix(1-((n-1):1)/(n+1), nrow=n-1, ncol=length(J)), as.data.frame(pseudodata2[,J]))
  res[,2] <- Cn.surv/Cn.survdenom
  return(res)
}

### Lower coefficient Frahm
lowerfrahm<- function(data){
  n <- nrow(data)
  d<- ncol(data)
  pseudodata<-pobs(data)
  res<-as.data.frame(matrix(NA,ncol=2, nrow=n-1)) # k=1, ..., n-1
  names(res)<-c("k", "lambda")
  res[,1]<-1:(n-1)
 
  Cn <- C.n(matrix((1:(n-1))/(n+1), nrow=n-1, ncol=d), pseudodata)
  pseudodata2 <- 1 - pseudodata + 1/n^2; # the last summand to obtain strict inequality in empirical survival function
  Cn.surv <- C.n(matrix(1-(1:(n-1))/(n+1), nrow=n-1, ncol=d), pseudodata2)
  res[,2] <- Cn/(1-Cn.surv);

  return(res)
}

### Upper coefficient Frahm
upperfrahm<- function(data){
  n <- nrow(data)
  d<- ncol(data)
  pseudodata<-pobs(data)
  res<-as.data.frame(matrix(NA,ncol=2, nrow=n-1))  # k=1, ..., n-1
  names(res)<-c("k", "lambda")
  res[,1]<-1:(n-1)
    
  Cn <- C.n(matrix(((n-1):1)/(n+1), nrow=n-1, ncol=d), pseudodata)
  pseudodata2 <- 1 - pseudodata + 1/n^2; # the last summand to obtain strict inequality in empirical survival function
  Cn.surv <- C.n(matrix(1-((n-1):1)/(n+1), nrow=n-1, ncol=d), pseudodata2)
  res[,2] <- Cn.surv/(1-Cn);

  return(res)
}


### Lower coefficient Schmid and Schmidt
lowerschmid <- function(data,estdenom=T){ # option T calculates the estimator given in the paper
  n <- nrow(data)
  d<- ncol(data)
  pseudodata<-pobs(data)
  res<-as.data.frame(matrix(NA,ncol=2, nrow=n-1)) # k=1, ..., n-1
  names(res)<-c("k", "lambda")
  res[,1]<-1:(n-1)

  pn <- 1:(n-1)/(n+1);
  temp <- pmax(outer(pn, pseudodata, "-"),0)  # positive part of the differences between pn and pseudobservations
  temp2 <- temp[,,1];    
  for(j in 2:d){ # product over dimensions
    temp2 <- temp2 * temp[,,j]
  }     

  denom <- rowSums(pmax(outer(pn, (1:n)/(n+1), "-"),0)^d);
  if(estdenom==T){res[,2] <- rowSums(temp2)/denom} else {res[,2] <- (d+1)/n*rowSums(temp2)*((n+1)/(1:(n-1)))^(d+1)}

  return(res)
}


### Upper coefficient Schmid and Schmidt
upperschmid <- function(data,estdenom=T){
  n <- nrow(data)
  d<- ncol(data)
  pseudodata<-pobs(data)
  res<-as.data.frame(matrix(NA,ncol=2, nrow=n-1))
  names(res)<-c("k", "lambda")
  res[,1]<-1:(n-1)
 
  pn <- 1:(n-1)/(n+1);
  temp <- pmax(outer(pn, 1 - pseudodata, "-"),0)  # positive part of the differences between pn and pseudobservations
  temp2 <- temp[,,1];    
  for(j in 2:d){ # product over dimensions
    temp2 <- temp2 * temp[,,j]
  }     

  denom <- rowSums(pmax(outer(pn, (1:n)/(n+1), "-"),0)^d);
  if(estdenom==T){res[,2] <- rowSums(temp2)/denom} else {res[,2] <- (d+1)/n*rowSums(temp2)*((n+1)/(1:(n-1)))^(d+1)}
  
  return(res)
  
}


### Madogram estimator of Pickands function
madesti<- function(w, data){
  n <- nrow(data)
  d<- ncol(data)
  pseudodata<-pobs(data)
  nu<-mean(apply(pseudodata,1, function(x){max(x^(1/w))-mean(x^(1/w))}))
  c<- mean(w/(1+w))
  return((nu+c)/(1-nu-c))
}


### Estimator of the coefficient for EV copulas
EVmeasure<- function(data){
  n <- nrow(data)
  d<- ncol(data)
  return((d/(d-1))*(1-madesti(rep(1/d, d),data)))
  
}

### Function to calculate all subsets of a given set
all.subsets.fast <- function(set) {
  n <- length(set)
  bin <- vector(mode = "list", length = n)
  for (i in 1L:n) {
    bin[[i]] <- rep.int(c(rep.int(F, 2L ^ (i - 1L)),
                          rep.int(T, 2L ^ (i - 1L))),
                        2L ^ (n - i))
  }
  apply(do.call(cbind, bin), 1L, function(x) { set[x] } )
}


### Upper Frahm coefficient using Pickands function
upperfrahmEV<- function(data){
  n <- nrow(data)
  d<- ncol(data)
  subs <- all.subsets.fast(c(1:ncol(data)))
  counter<-0
  for (i in 1:length(subs)){
    j<-length(subs[[i]]) 
    w<-rep(0,d)
    w[subs[[i]]]<-1/j
    counter<-counter+(-1)^(j+1)*j*madesti(w, data)
  }
  return(counter/(d*madesti(rep(1/d,d),data)))
}




##### Data loading and manipulation

# assumes working directory to contain the source csv files

files<- list.files(pattern="*.csv")
alldata <- read_csv(files[1])[,c("Date", "Adj Close")]
colnames(alldata)[2]<-substr(files[1],1,nchar(files[1])-4)
for (i in 2:length(files)){
  alldata<-alldata %>% full_join(read_csv(files[i])[,c("Date", "Adj Close")], by="Date")
  colnames(alldata)[i+1]<-substr(files[i],1,nchar(files[i])-4)
  
}
alldata<-as.data.frame(alldata)


# sort by date
alldata <- arrange(alldata, Date)


## use 15 years of data
alldata[1,1] ### first day
alldata[nrow(alldata),1] ### last day
dim(alldata)
which(alldata[,1]=="2005-03-18")

alldata<-alldata[which(alldata[,1]=="2005-03-18"):nrow(alldata), ]
dim(alldata) # 3848 observations



rownames(alldata)<-(alldata[,1])
alldata<-alldata[,-1]
alldata[]<-apply(alldata,2,as.numeric) # warning irrelevant
#alldata<-alldata[complete.cases(alldata),]
typeof(alldata)

alldata<- na.fill(alldata, fill = "extend") # fill the NAs by linear interpolation



apply(is.na(alldata), 2, which) # no NAs in the data


alldatareturn<-apply(alldata,2, function(x){-diff(log(x))}) # negative log returns
dim(alldatareturn) # one less observation because of the differences


fit<- fit_ARMA_GARCH(alldatareturn) # fitting univariate ARMA-GARCH 
Z <- as.data.frame(sapply(fit$fit, residuals, standardize = TRUE)) # standardized residuals from the fits
dim(Z)
dim(alldata)
colnames(Z)<-colnames(alldata)
rownames(Z)<-rownames(alldata)[-1]




### three stock groups
g1<-c("BAYN.DE", "DPW.DE", "BMW.DE") # German
g2<-c("BBVA.MC","BNP.PA", "G.MI") # financial
g3<-c("IBE.MC", "ENEL.MI", "ENGI.PA") # energetics



## group 1 only
data<-Z[,g1]

s1resfrahm<-upperfrahm(data)
s1resfrahmEV<-upperfrahmEV(data)
s1resschmid<-upperschmid(data)
s1resli1<-upperli(data,c(1))
s1resli2<-upperli(data,c("DPW.DE", "BMW.DE"))
s1resEV <- EVmeasure(data)


## group 2 only
data<-Z[,g2]

s2resfrahm<-upperfrahm(data)
s2resfrahmEV<-upperfrahmEV(data)
s2resschmid<-upperschmid(data)
s2resli1<-upperli(data,c(1))
s2resli2<-upperli(data,c("BBVA.MC", "G.MI"))
s2resEV <- EVmeasure(data)



## group 3 only
data<-Z[,g3]

s3resfrahm<-upperfrahm(data)
s3resfrahmEV<-upperfrahmEV(data)
s3resschmid<-upperschmid(data)
s3resli1<-upperli(data,c(1))
s3resli2<-upperli(data,c("IBE.MC",  "ENGI.PA"))
s3resEV <- EVmeasure(data)


## groups 1 and 2
data<-Z[,c(g1,g2)]

s4resfrahm<-upperfrahm(data)
s4resfrahmEV<-upperfrahmEV(data)
s4resschmid<-upperschmid(data)
s4resli1<-upperli(data,c(1))
s4resli2<-upperli(data,c("DPW.DE", "BMW.DE", "BBVA.MC","BNP.PA", "G.MI"))
s4resEV <- EVmeasure(data)


## groups 1 and 3
data<-Z[,c(g1,g3)]

s5resfrahm<-upperfrahm(data)
s5resfrahmEV<-upperfrahmEV(data)
s5resschmid<-upperschmid(data)
s5resli1<-upperli(data,c(1))
s5resli2<-upperli(data,c("BAYN.DE", "DPW.DE", "BMW.DE","IBE.MC", "ENGI.PA"))
s5resEV <- EVmeasure(data)


## groups 2 and 3
data<-Z[,c(g2,g3)]

s6resfrahm<-upperfrahm(data)
s6resfrahmEV<-upperfrahmEV(data)
s6resschmid<-upperschmid(data)
s6resli1<-upperli(data,c(1))
s6resli2<-upperli(data,c("BBVA.MC","BNP.PA", "G.MI","IBE.MC",  "ENGI.PA"))
s6resEV <- EVmeasure(data)



## all the three groups
data<-Z[,c(g1,g2,g3)]

s7resfrahm<-upperfrahm(data)
s7resfrahmEV<-upperfrahmEV(data)
s7resschmid<-upperschmid(data)
s7resli1<-upperli(data,c(1,2))
s7resli2<-upperli(data,c(1))
s7resEV <- EVmeasure(data)



###### Graphical settings

#dev.off()

linetypes<-c(1,2,3)
# colors = gray.colors(3, start = 0, end = 0.8, gamma=1)
#colors = c("darkgreen", "red", "blue")
colors<- c("black", "black", "black")
kmax<-500
plotheight<-3.7 # in inches
plotwidth<-6 # in inches
par(cex=1.7, lwd=2,mgp=c(2.2,0.45,0), tcl=-0.4, mar=c(3.3,3.6,1.1,1.1))
mypar<-par()
cexpar<-1.2




### Plotting Frahm

pdf("RD1.pdf", width=plotwidth, height = plotheight)
par(mypar)
par(cex=cexpar)


plot(s1resfrahm, main="", type="n", ylim=c(0,0.3), xlim=c(0,kmax), ylab="Tail coefficient estimate", yaxt="n")
axis(2, at=c(0.05, 0.1, 0.15, 0.2, 0.25, 0.3),labels=c(NA, 0.1, NA, 0.2, NA, 0.3))
lines(s1resfrahm, col=colors[1], lty=linetypes[1])
lines(s2resfrahm, col=colors[2], lty=linetypes[2])
lines(s3resfrahm, col=colors[3], lty=linetypes[3])
points(-10,s1resfrahmEV, col=colors[1], pch=15)
points(-10,s2resfrahmEV, col=colors[2], pch=16)
points(-10,s3resfrahmEV, col=colors[3], pch=17)

legend("bottomright",
       legend=c("G1", "G2", "G3"),  lty=linetypes, pch=c(15,16,17),  col=colors, merge=F)

#leg <- legend("bottomright", legend=c("G1", "G2", "G3"),   lty=linetypes, col=colors, plot = T)

#legend(x=c(leg$rect$left, leg$rect$left+leg$rect$w), y = c(leg$rect$top, leg$rect$top-leg$rect$h),
#       legend=c("G1", "G2", "G3"),  lty=linetypes, pch=c(15,16,17),  col=colors, merge=F)

#legend("bottomright",ncol = 2,
#       legend=c("","","","G1", "G2", "G3"),  lty=c(NA, NA, NA, linetypes), pch=c(15,16,17, NA, NA, NA),  col=colors)

dev.off()



pdf("RD2.pdf", width=plotwidth, height = plotheight)
par(mypar)
par(cex=cexpar)


plot(s1resfrahm, main="", type="n", ylim=c(0,0.3), xlim=c(0,kmax), ylab="Tail coefficient estimate", yaxt="n")
axis(2, at=c(0.05, 0.1, 0.15, 0.2, 0.25, 0.3),labels=c(NA, 0.1, NA, 0.2, NA, 0.3))
lines(s4resfrahm, col=colors[1], lty=linetypes[1])
lines(s5resfrahm, col=colors[2], lty=linetypes[2])
lines(s6resfrahm, col=colors[3], lty=linetypes[3])
points(-10,s4resfrahmEV, col=colors[1], pch=15)
points(-10,s5resfrahmEV, col=colors[2], pch=16)
points(-10,s6resfrahmEV, col=colors[3], pch=17)
legend("topright", legend=c("G1 + G2", "G1 + G3", "G2 + G3"), lty=linetypes, col=colors, pch=c(15,16,17), merge = F)


dev.off()

#lines(s7resfrahm, col=colors[1], lty=linetypes[3])




####### plotting Schmid and Schmidt

pdf("RD3.pdf", width=plotwidth, height = plotheight)
par(mypar)
par(cex=cexpar)


plot(s1resschmid, main="", type="n", ylim=c(0,0.7), xlim=c(0,kmax), ylab="Tail coefficient estimate")
lines(s1resschmid, col=colors[1], lty=linetypes[1])
lines(s2resschmid, col=colors[2], lty=linetypes[2])
lines(s3resschmid, col=colors[3], lty=linetypes[3])
legend("bottomright", legend=c("G1", "G2", "G3"), lty=linetypes, col=colors)


dev.off()



pdf("RD4.pdf", width=plotwidth, height = plotheight)
par(mypar)
par(cex=cexpar)



plot(s1resschmid, main="", type="n", ylim=c(0,0.7), xlim=c(0,kmax), ylab="Tail coefficient estimate")
lines(s4resschmid, col=colors[1], lty=linetypes[1])
lines(s5resschmid, col=colors[2], lty=linetypes[2])
lines(s6resschmid, col=colors[3], lty=linetypes[3])
legend("topright", legend=c("G1 + G2", "G1 + G3", "G2 + G3"), lty=linetypes, col=colors)

dev.off()




#### plotting Li

pdf("RD5.pdf", width=plotwidth, height = plotheight)
par(mypar)
par(cex=cexpar)


plot(s1resli1, main="", type="n", ylim=c(0,0.6), xlim=c(0,kmax),
     ylab="Tail coefficient estimate")
lines(s1resli1, col=colors[1], lty=linetypes[1])
lines(s2resli1, col=colors[2], lty=linetypes[2])
lines(s3resli1, col=colors[3], lty=linetypes[3])
legend("bottomright", legend=c("G1", "G2", "G3"), lty=linetypes, col=colors)


dev.off()


pdf("RD6.pdf", width=plotwidth, height = plotheight)
par(mypar)
par(cex=cexpar)


plot(s1resli1, main="", type="n", ylim=c(0,0.6), xlim=c(0,kmax),
     ylab="Tail coefficient estimate")
lines(s4resli1, col=colors[1], lty=linetypes[1])
lines(s5resli1, col=colors[2], lty=linetypes[2])
lines(s6resli1, col=colors[3], lty=linetypes[3])
legend("topright", legend=c("G1 + G2", "G1 + G3", "G2 + G3"), lty=linetypes, col=colors)

dev.off()



pdf("RD7.pdf", width=plotwidth, height = plotheight)
par(mypar)
par(cex=cexpar)


plot(s1resli2, main="", type="n", ylim=c(0,1), xlim=c(0,kmax),
     ylab="Tail coefficient estimate")
lines(s1resli2, col=colors[1], lty=linetypes[1])
lines(s2resli2, col=colors[2], lty=linetypes[2])
lines(s3resli2, col=colors[3], lty=linetypes[3])
legend("bottomright", legend=c("G1", "G2", "G3"), lty=linetypes, col=colors)


dev.off()


pdf("RD8.pdf", width=plotwidth, height = plotheight)
par(mypar)
par(cex=cexpar)



plot(s1resli2, main="", type="n", ylim=c(0,1), xlim=c(0,kmax), 
     ylab="Tail coefficient estimate")
lines(s4resli2, col=colors[1], lty=linetypes[1])
lines(s5resli2, col=colors[2], lty=linetypes[2])
lines(s6resli2, col=colors[3], lty=linetypes[3])
legend("bottomright", legend=c("G1 + G2", "G1 + G3", "G2 + G3"), lty=linetypes, col=colors)

dev.off()





#### EV measure results
paste0(round(c(s1resEV, s2resEV, s3resEV, s4resEV, s5resEV, s6resEV),2), collapse="$ & $" )

